import 'package:fast_base58/fast_base58.dart';
import 'package:hex/hex.dart';
import 'package:bip32_ed25519/api.dart';
import 'package:bip39_multi_nullsafety/bip39_multi_nullsafety.dart' as bip39;
import 'package:pinenacl/key_derivation.dart';

class HdWallet {
  final Bip32SigningKey rootSigningKey;
  final _derivator = Bip32Ed25519KeyDerivation.instance;

  /// Root constructor taking a root signing key
  HdWallet({required this.rootSigningKey});

  factory HdWallet.fromHexEntropy(String hexEntropy) => HdWallet(
        rootSigningKey: _bip32signingKey(
          Uint8List.fromList(
            HEX.decode(hexEntropy),
          ),
        ),
      );

  /// Build the root signing key based on given mnemonic.
  factory HdWallet.fromMnemonic(String mnemonic, {String lang = 'english'}) =>
      HdWallet.fromHexEntropy(
        bip39.mnemonicToEntropy(mnemonic.trim(), language: lang),
      );

  /// return the root signing key
  Bip32VerifyKey get rootVerifyKey => rootSigningKey.verifyKey;

  /// derive root signing key given a seed
  static Bip32SigningKey _bip32signingKey(Uint8List seed) {
    final rawMaster = PBKDF2.hmac_sha512(
        Uint8List(0), seed, 4096, cip16ExtendedSigningKeySize);
    final Bip32SigningKey rootXsk = Bip32SigningKey.normalizeBytes(rawMaster);
    return rootXsk;
  }

  /// If a parent signing key is provided, a child signing key is generated. If a parent
  /// verify key is provided and the index is NOT hardened, then a child verify key is
  /// also included. If hardened and no signingKey is provied, it returns an empty pair.
  Bip32KeyPair derive({required Bip32KeyPair keys, required int index}) {
    // computes a child extended private key from the parent extended private key.
    Bip32SigningKey? signingKey = keys.signingKey != null
        ? _derivator.ckdPriv(keys.signingKey!, index) as Bip32SigningKey
        : null;
    Bip32VerifyKey? verifyKey = isHardened(index)
        ? null
        : keys.verifyKey != null
            ? _derivator.ckdPub(keys.verifyKey!, index) as Bip32VerifyKey
            : _derivator.neuterPriv(signingKey!) as Bip32VerifyKey;
    return Bip32KeyPair(signingKey: signingKey, verifyKey: verifyKey);
  }

  /// Provide a valid signature from given derivation keypair for any message or text document.
  SignatureBase sign(String document, {required Bip32KeyPair keypair}) {
    return keypair.signingKey!.sign(document.codeUnits).signature;
  }

  /// Verify if the given signature is valid for the given message or text document,
  /// based on given keypair
  bool verifySign(String document, SignatureBase signature,
      {required Bip32KeyPair keypair}) {
    return keypair.verifyKey!.verify(
        signature: signature, message: document.codeUnits.toUint8List());
  }

  /// Return the pubkey in string format for the given derivation keypair
  String getPubkey(Bip32KeyPair derivation) {
    return Base58Encode(derivation.signingKey!.publicKey.rawKey.toList());
  }
}

/// Private/signing and public/varification key pair.
class Bip32KeyPair {
  final Bip32SigningKey? signingKey;
  final Bip32VerifyKey? verifyKey;
  const Bip32KeyPair({this.signingKey, this.verifyKey});
}

/// Return a 12 random worlds mnemonic based on bip39 standard.
/// You can choose a custom language like french, default is english.
String generateMnemonic({String lang = 'english'}) {
  return bip39.generateMnemonic(strength: 128, language: lang);
}

/// Hardended chain values should not have public keys.
/// They are denoted by a single quote in chain values.
const int hardenedOffset = 0x80000000;

/// Extended private key size in bytes
const cip16ExtendedSigningKeySize = 96;

/// Hardens index, meaning it won't have a public key
int harden(int index) => index | hardenedOffset;

/// Returns true if index is hardened.
bool isHardened(int index) => index & hardenedOffset != 0;
